/*-------------------------------------------------------------------------+
| Copyright 2022 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.utils;

import static java.time.format.DateTimeFormatter.ISO_DATE_TIME;
import static java.time.format.DateTimeFormatter.RFC_1123_DATE_TIME;
import static org.fortiss.tooling.ext.reuse.service.ReuseProviderService.getPossibleExternalReferenceClasses;
import static org.fortiss.tooling.ext.reuse.service.ReuseProviderService.getReuseElementClasses;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.getAllSeparatelyStoredReuseLibraries;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.ext.reuse.model.ReuseElementSpec;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.service.IReuseProviderService;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * Utility methods for {@link ReuseLibrary}s and related reuse elements and
 * actions with basics like getters and setters.
 *
 * @author bergemann
 */
public class ReuseLibraryUtilsBasics {

	/** Standard internal date format used for reuse (e.g. for hash dates) */
	public final static DateTimeFormatter INTERNAL_REUSE_DATE_FORMATTER = ISO_DATE_TIME;

	/** Standard date format used for reuse (easier readable) */
	public final static DateTimeFormatter DISPLAY_REUSE_DATE_FORMATTER = RFC_1123_DATE_TIME;

	/**
	 * Returns whether the given element can be used as reuse element.
	 *
	 * @param element The element that should be checked
	 * @return True if the given element can be used as reuse element otherwise
	 *         false
	 */
	public static boolean hasReusePossibility(EObject element) {
		for (Class<? extends EObject> cls : getReuseElementClasses()) {
			if (cls.isInstance(element)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns whether the given element is a reuse element and has at least one
	 * valid link to a {@link ReuseLibrary}.
	 *
	 * @param element The element that should be checked
	 * @return True if the given element is a reuse element, otherwise false
	 */
	public static boolean isCorrectReuseElement(EObject element) {
		// first, check if it is a (possible) reuse element in general
		if (hasReusePossibility(element)) {
			// second, check if it has a library reference
			// (only then it is an actual (active) reuse element)
			if (element instanceof IModelElement) {
				for (IModelElementSpecification spec : ((IModelElement) element).getSpecifications()) {
					if (spec instanceof ReuseElementSpec) {
						String libID = ((ReuseElementSpec) spec).getSourceLibUUID();
						if (libID != null && !libID.isEmpty()) {
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * Returns whether the given element is a reuse element and has at least one
	 * valid link to a {@link ReuseLibrary} OR is a {@link ReuseLibrary} itself.
	 *
	 * @param element The element that should be checked
	 * @return True if the given element is a reuse element OR a reuse library,
	 *         otherwise false
	 */
	public static boolean isCorrectReuseElementOrLibrary(EObject element) {
		return (element instanceof ReuseLibrary) || isCorrectReuseElement(element);
	}

	/**
	 * Returns display name of given {@link ReuseLibrary} (name and UUID).
	 *
	 * @param library The target reuse library
	 * @return Display name as string
	 */
	public static String getLibraryIDName(ReuseLibrary library) {
		return getLibraryIDNameFormat(library.getLibraryUUID(), library.getName());
	}

	/**
	 * Returns the current format for displaying name and UUID of
	 * {@link ReuseLibrary}s.
	 *
	 * @param libraryUUID The UUID of the target library
	 * @param libraryName The name of the target library
	 * @return Formatted string
	 */
	public static String getLibraryIDNameFormat(String libraryUUID, String libraryName) {
		return libraryName + " (" + libraryUUID + ")";
	}

	/**
	 * Returns the (converted) name of the given reuse element (or
	 * {@link ReuseLibrary}), independent of the type as long as its class was
	 * registered for reuse.
	 *
	 * @param element The target reuse element
	 * @return Name as string
	 */
	public static String getReuseElementName(EObject element) {
		if (element instanceof ReuseLibrary) {
			return ((ReuseLibrary) element).getName();
		}
		String name = IReuseProviderService.getInstance().getReuseElementName(element);
		if (name == null) {
			if (element instanceof INamedElement) {
				name = ((INamedElement) element).getName();
			} else {
				name = element.toString();
			}
		}
		return name;
	}

	/**
	 * Sets the name of the given reuse element (or {@link ReuseLibrary}),
	 * independent of the type as long as its class was registered for reuse.
	 *
	 * @param element The target reuse element
	 * @param name    The new name
	 * @return True if action was successful, otherwise false
	 */
	public static boolean setReuseElementName(EObject element, String name) {
		if (element instanceof ReuseLibrary) {
			((ReuseLibrary) element).setName(name);
			return true;
		}
		return IReuseProviderService.getInstance().setReuseElementName(element, name);
	}

	/**
	 * Returns the comment of the given reuse element (or {@link ReuseLibrary}),
	 * independent of the type as long as its class was registered for reuse.
	 *
	 * @param element The target reuse element
	 * @return The requested comment as string
	 */
	public static String getReuseElementComment(EObject element) {
		String comment = null;
		if (element instanceof ReuseLibrary) {
			comment = ((ReuseLibrary) element).getComment();
		} else {
			comment = IReuseProviderService.getInstance().getReuseElementComment(element);
			if (comment == null) {
				if (element instanceof INamedCommentedElement) {
					comment = ((INamedCommentedElement) element).getComment();
				}
			}
		}
		return (comment == null) ? "" : comment;
	}

	/**
	 * Sets the comment of the given reuse element (or {@link ReuseLibrary}),
	 * independent of the type as long as its class was registered for reuse.
	 *
	 * @param element The target reuse element
	 * @param comment The new comment
	 * @return True if action was successful, otherwise false
	 */
	public static boolean setReuseElementComment(EObject element, String comment) {
		if (element instanceof ReuseLibrary) {
			((ReuseLibrary) element).setComment(comment);
			return true;
		}
		return IReuseProviderService.getInstance().setReuseElementComment(element, comment);
	}

	/**
	 * Returns the first UUID of the given reuse element (or {@link ReuseLibrary})
	 * as long as it has (already) a {@link ReuseElementSpec}. If this is not the
	 * case, an empty {@link String} will be returned. This method is perfect for
	 * reuse elements within reuse libraries, because they have definitely only one
	 * {@link ReuseElementSpec} and thus exactly one UUID (= the first one). Reuse
	 * elements that are already placed in projects can have multiple
	 * {@link ReuseElementSpec}s and thus multiple UUIDs, because they could have
	 * been added to several {@link ReuseLibrary}s.
	 *
	 * @param element The target reuse element
	 * @return The requested UUID as string (or empty string)
	 */
	public static String getFirstReuseElementUUID(EObject element) {
		List<String> allUUIDs = getAllReuseElementUUID(element);
		if (allUUIDs != null && !allUUIDs.isEmpty()) {
			// get(0) is safe due to check above
			return allUUIDs.get(0);
		}
		return "";
	}

	/**
	 * Returns all UUIDs of the given reuse element (or {@link ReuseLibrary}) as
	 * long as it has (already) at least one {@link ReuseElementSpec}. If this is
	 * not the case, an empty list will be returned. This method exists because
	 * reuse elements that are already placed in projects can have multiple
	 * {@link ReuseElementSpec}s and thus multiple UUIDs, because they could have
	 * been added to several {@link ReuseLibrary}s.
	 *
	 * @param element The target reuse element
	 * @return The requested UUID as string (or empty string)
	 */
	public static List<String> getAllReuseElementUUID(EObject element) {
		List<String> allUUIDs = new ArrayList<>();
		if (element instanceof ReuseLibrary) {
			allUUIDs.add(((ReuseLibrary) element).getLibraryUUID());
		} else {
			if (element instanceof IModelElement) {
				List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
				for (ReuseElementSpec spec : specs) {
					allUUIDs.add(spec.getElementUUID());
				}
			}
		}
		return allUUIDs;
	}

	/**
	 * Returns the UUID of the given reuse element that is stored in the
	 * {@link ReuseElementSpec} which is connected to the given
	 * {@link ReuseLibrary}. If such a {@link ReuseElementSpec} does not exist, an
	 * empty {@link String} will be returned.
	 *
	 * @param element The target reuse element
	 * @param library The reuse library to which the reuse UUID of the element
	 *                should belong
	 * @return The requested UUID as string (or empty string)
	 */
	public static String getReuseElementUUIDForLibrary(EObject element, ReuseLibrary library) {
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
			for (ReuseElementSpec spec : specs) {
				if (spec.getSourceLibUUID().equals(library.getLibraryUUID())) {
					return spec.getElementUUID();
				}
			}
		}
		return "";
	}

	/**
	 * Returns the hash code string of the given reuse element for its origin inside
	 * a specific (given) {@link ReuseLibrary}. If it cannot be found, an empty
	 * {@link String} will be returned.
	 *
	 * @param element       The target reuse element
	 * @param sourceLibrary The library of the original reuse element for the hash
	 * @return The requested hash code as string (or empty string)
	 */
	public static String getReuseElementHashForLibrary(EObject element, ReuseLibrary sourceLibrary) {
		String libraryUuid = sourceLibrary.getLibraryUUID();
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
			for (ReuseElementSpec spec : specs) {
				String referenceLibraryUuid = spec.getSourceLibUUID();
				if (referenceLibraryUuid.equals(libraryUuid)) {
					return spec.getElementHash();
				}
			}
		}
		return "";
	}

	/**
	 * Returns the Last Update Date of the given reuse element as
	 * {@link ZonedDateTime} as long as it has (already) a {@link ReuseElementSpec}.
	 * If this is not the case, null will be returned.
	 *
	 * @param element The target reuse element
	 * @return The requested (update) date (or null)
	 */
	public static ZonedDateTime getReuseElementLastUpdate(EObject element) {
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs((IModelElement) element);
			if (specs != null && !specs.isEmpty()) {
				// get(0) is safe due to check above
				return specs.get(0).getLastUpdate();
			}
		}
		return null;
	}

	/**
	 * Returns the Last Update Date of the given reuse element as formatted
	 * {@link String} as long as it has (already) a {@link ReuseElementSpec}. If
	 * this is not the case, an empty {@link String} will be returned.
	 *
	 * @param element    The target reuse element
	 * @param forDisplay If it should be the format for user display or the one that
	 *                   is internally used
	 * @return The requested (update) date as string (or empty string)
	 */
	public static String getReuseElementLastUpdateString(EObject element, boolean forDisplay) {
		ZonedDateTime updateDate = getReuseElementLastUpdate(element);
		return getDateAsString(updateDate, forDisplay);
	}

	/**
	 * Returns the given {@link ZonedDateTime} as formatted {@link String} as long
	 * as it is valid, otherwise an empty {@link String} will be returned.
	 *
	 * @param date       The date
	 * @param forDisplay If it should be the format for user display or the one that
	 *                   is internally used
	 * @return The requested date as string (or empty string)
	 */
	public static String getDateAsString(ZonedDateTime date, boolean forDisplay) {
		if (date != null) {
			if (forDisplay) {
				return DISPLAY_REUSE_DATE_FORMATTER.format(date);
			} else {
				return INTERNAL_REUSE_DATE_FORMATTER.format(date);
			}
		}
		return "";
	}

	/**
	 * Sets the Last Update Date of the given reuse element with the given {link
	 * Date} as long as it has (already) a {@link ReuseElementSpec}.
	 *
	 * @param element The target reuse element
	 * @param newDate The new (update) date
	 */
	public static void setReuseElementLastUpdate(EObject element, ZonedDateTime newDate) {
		if (element instanceof IModelElement) {
			for (ReuseElementSpec spec : getAllReuseSpecs((IModelElement) element)) {
				spec.setLastUpdate(newDate);
			}
		}
	}

	/**
	 * Sets the Last Update Date of the given reuse element with the current {link
	 * ZonedDateTime} as long as it has (already) a {@link ReuseElementSpec}.
	 *
	 * @param element The target reuse element
	 */
	public static void setReuseElementLastUpdateNow(EObject element) {
		setReuseElementLastUpdate(element, ZonedDateTime.now());
	}

	/**
	 * Returns all {@link ReuseElementSpec}s of the given element to a
	 * {@link ReuseLibrary}. If none exist, an empty list will be returned.
	 *
	 * @param element The target reuse element
	 * @return A list of all found reuse specifications (or empty list)
	 */
	public static List<ReuseElementSpec> getAllReuseSpecs(IModelElement element) {
		List<ReuseElementSpec> specs = new ArrayList<>();
		for (IModelElementSpecification spec : element.getSpecifications()) {
			if (spec instanceof ReuseElementSpec) {
				// (cast is safe due to instance check)
				specs.add((ReuseElementSpec) spec);
			}
		}
		return specs;
	}

	/**
	 * Returns the number of all {@link ReuseElementSpec}s of the given element.
	 *
	 * @param element The target reuse element
	 * @return Number of all found reuse specifications
	 */
	public static int getNumberOfContainedReuseSpecs(EObject element) {
		if (element instanceof IModelElement) {
			return getAllReuseSpecs((IModelElement) element).size();
		} else {
			return 0;
		}
	}

	/**
	 * Returns the {@link ReuseElementSpec} of the given (project) reuse element
	 * that is linked with the given reuse element inside a library. If none could
	 * be found, null will be returned.
	 *
	 * @param projectElement The reuse element inside a project from which the reuse
	 *                       specification should be selected
	 * @param libraryElement The reuse element inside a library
	 * @return The found/matched reuse specification (or null)
	 */
	public static ReuseElementSpec getReuseSpecLinkedToThisLibraryElement(IModelElement projectElement,
			IModelElement libraryElement) {
		List<ReuseElementSpec> projectSpecs = getAllReuseSpecs(projectElement);
		List<ReuseElementSpec> librarySpecs = getAllReuseSpecs(libraryElement);
		// elements in library must have only a single specification
		if (librarySpecs.size() == 1) {
			// get(0) is safe due to size check above
			ReuseElementSpec librarySpec = librarySpecs.get(0);

			for (ReuseElementSpec specToBeChecked : projectSpecs) {
				// both UUIDs (of element and library) must be identical
				if (specToBeChecked.getSourceLibUUID().equals(librarySpec.getSourceLibUUID())
						&& specToBeChecked.getElementUUID().equals(librarySpec.getElementUUID())) {
					return specToBeChecked;
				}
			}
		}
		return null;
	}

	/**
	 * Returns all {@link ReuseLibrary}s that are present in the local workspace.
	 * Will return empty list if no {@link ReuseLibrary} exists.
	 *
	 * @return A list of all local reuse libraries (or empty list)
	 */
	public static List<ReuseLibrary> getAllLocalReuseLibraries() {
		// currently, this is just all existing local libraries which are stored
		// separately in the library directory
		return getAllSeparatelyStoredReuseLibraries();
	}

	/**
	 * Returns the {@link ReuseLibrary} with the given UUID out of all local
	 * {@link ReuseLibrary}. Will return null if no {@link ReuseLibrary} exists with
	 * this UUID. Note: it will return the {@link ReuseLibrary} even if the folder
	 * or file name is not correct (as long as it is in the reuse directory and has
	 * the right UUID)!
	 *
	 * @param libraryUUID The UUID of the requested reuse library
	 * @return The requested reuse library (or null)
	 */
	public static ReuseLibrary getLocalReuseLibraryByID(String libraryUUID) {
		for (ReuseLibrary library : getAllLocalReuseLibraries()) {
			if (libraryUUID.equals(library.getLibraryUUID())) {
				return library;
			}
		}
		return null;
	}

	/**
	 * Returns the {@link ReuseLibrary} that is mentioned as source library in the
	 * first {link ReuseElementSpec} of the given reuse element. If no {link
	 * ReuseElementSpec} exists, the return will be null. This method is faster than
	 * calling getAllSourceLibrariesOfElement() and returning the first library,
	 * because it does not fetch all libraries - only the one of the first {link
	 * ReuseElementSpec}.
	 *
	 * @param element The target reuse element
	 * @return The requested reuse library (or null)
	 */
	public static ReuseLibrary getFirstSourceLibraryOfElement(EObject element) {
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> specList = getAllReuseSpecs((IModelElement) element);
			if (specList != null && !specList.isEmpty()) {
				// get(0) is safe due to check above
				String libraryUUID = specList.get(0).getSourceLibUUID();
				return getLocalReuseLibraryByID(libraryUUID);
			}
		}
		return null;
	}

	/**
	 * Returns a list of {@link ReuseLibrary}s that are each mentioned as source
	 * library in the {link ReuseElementSpec}s of the given reuse element. If no
	 * {link ReuseElementSpec} exists, the return will be an empty list.
	 *
	 * @param element The target reuse element
	 * @return All found reuse libraries for the given element (or empty list)
	 */
	public static List<ReuseLibrary> getAllSourceLibrariesOfElement(EObject element) {
		List<ReuseLibrary> foundLibraries = new ArrayList<>();
		if (element instanceof IModelElement) {
			List<ReuseElementSpec> reuseSpecs = getAllReuseSpecs((IModelElement) element);
			for (ReuseElementSpec reuseSpec : reuseSpecs) {
				String libraryUUID = reuseSpec.getSourceLibUUID();
				ReuseLibrary sourceLibrary = getLocalReuseLibraryByID(libraryUUID);
				if (sourceLibrary != null) {
					foundLibraries.add(sourceLibrary);
				}
			}
		}
		return foundLibraries;
	}

	/**
	 * Returns the element that exists in the given {@link ReuseLibrary} and is
	 * referenced by the given {@link ReuseElementSpec}. Returns null if it could
	 * not be found.
	 *
	 * @param library     The target reuse library
	 * @param elementUUID The UUID string of the requested element
	 * @return The requested reuse element (or null)
	 */
	public static EObject getElementInsideLibrary(ReuseLibrary library, String elementUUID) {
		// find original element inside the library by identical UUID
		for (EObject libraryElement : library.getReuseElementList()) {
			for (IModelElementSpecification spec : ((IModelElement) libraryElement).getSpecifications()) {
				if (spec instanceof ReuseElementSpec) {
					if (((ReuseElementSpec) spec).getElementUUID().equals(elementUUID)) {
						return libraryElement;
					}
				}
			}
		}
		return null;
	}

	/**
	 * Returns a list of external elements that are used within the given element
	 * (e.g. an external stored function inside a code specification of a
	 * component). If nothing was found, an empty list is returned.
	 *
	 * @param element The target reuse element
	 * @return A list of external elements referencing the target reuse element (or
	 *         empty list)
	 */
	public static List<EObject> getExternalReferencesOfElement(EObject element) {
		List<EObject> references = new ArrayList<>();
		for (Class<? extends EObject> referenceClass : getPossibleExternalReferenceClasses()) {
			for (EObject reference : getChildrenWithType(element, referenceClass)) {
				references.add(reference);
			}
		}
		return references;
	}

	/**
	 * Returns a list with all found origin {@link EObject}s of the given
	 * {@link EObject} that were in the meantime updated inside their referenced
	 * {@link ReuseLibrary}. If no reuse reference could be found or if some could
	 * be found but all of these origins have not changed until the given
	 * {@link EObject} was reused based on the origin (i.e., the hash codes are
	 * identical), an empty list will be returned.
	 *
	 * @param element The element whose reuse origin(s) should be checked for
	 *                updates
	 * @return List of elements that are reuse origins and were updated in the
	 *         meantime
	 */
	public static List<EObject> getAllUpdatedOrigins(EObject element) {
		List<EObject> updatedOrigins = new ArrayList<>();

		List<ReuseLibrary> foundLibraries = getAllSourceLibrariesOfElement(element);

		for (ReuseLibrary sourceLibrary : foundLibraries) {
			String uuid = getReuseElementUUIDForLibrary(element, sourceLibrary);
			EObject originElementInLib = getElementInsideLibrary(sourceLibrary, uuid);
			String storedHashForThisLibrary = getReuseElementHashForLibrary(element, sourceLibrary);
			String currentHashForThisLibrary = getReuseElementHashForLibrary(originElementInLib, sourceLibrary);
			// if hashes are valid but different, the origin was updated
			if (!storedHashForThisLibrary.isEmpty() && !currentHashForThisLibrary.isEmpty()) {
				if (!storedHashForThisLibrary.equals(currentHashForThisLibrary)) {
					updatedOrigins.add(originElementInLib);
				}
			}
		}
		return updatedOrigins;
	}
}
